import boto.sqs
from motorway.contrib.amazon_sqs.utils import SQSJSONMessage
from motorway.intersection import Intersection


class SQSInsertIntersection(Intersection):
    queue_name = None
    sqs_message_class = SQSJSONMessage

    def __init__(self, **kwargs):
        super(SQSInsertIntersection, self).__init__(**kwargs)
        conn = boto.sqs.connect_to_region(**self.connection_parameters())
        assert self.queue_name, "Please define attribute queue_name on your SQSRamp"
        self.queue = conn.create_queue(self.queue_name)

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def process(self, message):
        self.queue.write(
            self.sqs_message_class(body=message.content)
        )
        self.ack(message)
        yield



from unittest import TestCase
import datetime
import uuid
import zmq
from motorway.controller import ControllerIntersection
from motorway.messages import Message
from motorway.tests.sample_pipeline import SampleWordCountPipeline
from motorway.tests.utils import override_process_name, ZMQSockMock
from freezegun import freeze_time


class ControllerTestCase(TestCase):

    # Utilities to test the controller

    def setUp(self):
        ctx = zmq.Context()
        self.controller = ControllerIntersection()
        self.control_messages = []
        self.controller_sock = ZMQSockMock(self.control_messages)

    def get_control_messages(self):
        messages = []
        while self.control_messages:
            messages.append(Message.from_message(self.control_messages.pop(0), self.controller_sock))
        return messages

    def run_controller_process_method(self):
        return list(
            self.controller.process(self.get_control_messages())
        )  # Remember list(), to evaluate a generator

    # Tests

    def test_histogram_success(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        with freeze_time("2014-01-10 12:00:01"):
            msg_in_ramp = Message(1337, "split this string please", producer_uuid=ramp_uuid)
            msg_in_ramp.send_control_message(self.controller_sock, datetime.timedelta(seconds=2), process_name=ramp_uuid, destination_uuid=intersection_uuid)
            self.run_controller_process_method()
            self.controller.update()
            msg_in_intersection = Message.from_message(msg_in_ramp._message(), self.controller_sock)
            msg_in_intersection.ack()
            self.run_controller_process_method()
            self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['histogram'][0]['success_count'], 1)
        with freeze_time("2014-01-10 12:59:01"):
            self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['histogram'][0]['success_count'], 0)

    def test_histogram_error(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        with freeze_time("2014-01-10 12:00:01"):
            msg_in_ramp = Message(1337, "split this string please", producer_uuid=ramp_uuid)
            msg_in_ramp.send_control_message(self.controller_sock, datetime.timedelta(seconds=2), process_name=ramp_uuid, destination_uuid=intersection_uuid)
            self.run_controller_process_method()
            self.controller.update()
            msg_in_intersection = Message.from_message(msg_in_ramp._message(), self.controller_sock)
            msg_in_intersection.fail("some error message", False)
            self.run_controller_process_method()
            self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['histogram'][0]['error_count'], 1)
        with freeze_time("2014-01-10 12:59:01"):
            self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['histogram'][0]['error_count'], 0)

    def test_frequency(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        msg = Message(1337, "split this string please", producer_uuid=ramp_uuid)
        msg.send_control_message(self.controller_sock, datetime.timedelta(seconds=2), process_name=ramp_uuid, destination_uuid=intersection_uuid, sender='ramp')
        self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['frequency'][2], 1)

    def test_waiting(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        msg_in_ramp = Message(1337, "split this string please", producer_uuid=ramp_uuid)
        msg_in_ramp.send_control_message(self.controller_sock, datetime.timedelta(seconds=2), process_name=ramp_uuid, destination_uuid=intersection_uuid)
        self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['waiting'], 0)
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['waiting'], 1)
        msg_in_intersection = Message.from_message(msg_in_ramp._message(), self.controller_sock, process_name=intersection_uuid)
        msg_in_intersection.ack()
        self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['waiting'], 0)
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['waiting'], 0)

    def test_message_timeout(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        with freeze_time("2014-01-10 12:00:01"):
            msg_in_ramp = Message(1337, "split this string please", producer_uuid=ramp_uuid)
            msg_in_ramp.send_control_message(self.controller_sock, datetime.timedelta(seconds=2), process_name=ramp_uuid, destination_uuid=intersection_uuid)
            self.run_controller_process_method()
            self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['waiting'], 0)
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['waiting'], 1)
        with freeze_time("2014-01-10 12:59:01"):
            self.controller.update()
            self.assertEqual(self.controller.process_statistics[ramp_uuid]['waiting'], 0)
            self.assertEqual(self.controller.process_statistics[intersection_uuid]['waiting'], 0)
            self.run_controller_process_method()
            self.assertEqual(self.controller.process_statistics[ramp_uuid]['histogram'][59]['timeout_count'], 0)
            self.assertEqual(self.controller.process_statistics[intersection_uuid]['histogram'][59]['timeout_count'], 1)

    def test_message_duration(self):
        ramp_uuid = str(uuid.uuid4())
        intersection_uuid = str(uuid.uuid4())
        intersection_two_uuid = str(uuid.uuid4())

        msg_in_ramp = Message(1337, "split this string please", producer_uuid=ramp_uuid)
        msg_in_ramp.send_control_message(self.controller_sock, datetime.timedelta(seconds=30), process_name=ramp_uuid, destination_uuid=intersection_uuid, sender='ramp')

        self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['avg_time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['total_frequency'], 1)
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['time_taken'], datetime.timedelta(seconds=0))
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['avg_time_taken'], datetime.timedelta(seconds=0))
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['total_frequency'], 0)


        msg_in_intersection = Message.from_message(msg_in_ramp._message(), self.controller_sock, process_name=intersection_uuid)
        new_msg_in_intersection = Message(1338, "yo", producer_uuid=intersection_uuid)
        new_msg_in_intersection.send_control_message(self.controller_sock, datetime.timedelta(seconds=7), process_name=intersection_uuid, destination_uuid=intersection_uuid)
        msg_in_intersection.ack(time_consumed=datetime.timedelta(seconds=15))

        self.run_controller_process_method()
        # self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['avg_time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['total_frequency'], 1)
        # self.assertEqual(self.controller.process_statistics[intersection_uuid]['time_taken'], datetime.timedelta(seconds=15))
        # self.assertEqual(self.controller.process_statistics[intersection_uuid]['avg_time_taken'], datetime.timedelta(seconds=15))
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['total_frequency'], 1)

        msg_in_intersection_two = Message.from_message(msg_in_ramp._message(), self.controller_sock, process_name=intersection_two_uuid)
        msg_in_intersection_two.ack()

        self.run_controller_process_method()
        self.controller.update()
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['avg_time_taken'], datetime.timedelta(seconds=30))
        self.assertEqual(self.controller.process_statistics[ramp_uuid]['total_frequency'], 1)
        # self.assertEqual(self.controller.process_statistics[intersection_uuid]['time_taken'], datetime.timedelta(seconds=15))
        # self.assertEqual(self.controller.process_statistics[intersection_uuid]['avg_time_taken'], datetime.timedelta(seconds=15))
        self.assertEqual(self.controller.process_statistics[intersection_uuid]['total_frequency'], 1)






import logging
import traceback

logger = logging.getLogger(__name__)

def batch_process(wait=5, limit=100):
    def inner(method):
        method.batch_process = True
        method.wait = wait
        method.limit = limit
        return method
    return inner


def post_error_to_sentry(client):
    def inner(method):
        method.post_error_to_sentry = True
        return method
    return inner
    # def inner(method):
    #     def dec(*args, **kwargs):
    #         try:
    #             logger.info("Running method()")
    #             return method(*args, **kwargs)
    #         except:
    #             logger.error("Caught exception", exc_info=True)
    #             print "this happened"
    #             identifier = client.get_ident(client.captureException())
    #             # print traceback.format_exc()
    #             # print "Exception caught; reference is %s" % identifier
    #             # raise
    #     # print method, dec
    #     return dec
    # return inner


import datetime

import itertools
import random
import socket
import uuid

import zmq
import time

from motorway.exceptions import SocketBlockedException
from motorway.grouping import HashRingGrouper, RandomGrouper, GroupingValueMissing, SendToAllGrouper
from motorway.messages import Message
from motorway.utils import set_timeouts_on_socket, get_connections_block, get_ip


class GrouperMixin(object):
    def get_grouper(self, grouper_name):
        if grouper_name == "HashRingGrouper":
            return HashRingGrouper
        elif grouper_name == "SendToAllGrouper":
            return SendToAllGrouper
        else:
            return RandomGrouper


class SendMessageMixin(object):

    def send_message(self, message, process_id, time_consumed=None, sender=None, control_message=True, retries=100):
        """

        :param message: A ::pipeline.messages.Message:: instance
        :param process_id:
        :param time_consumed:
        :param sender:
        :param control_message:
        :param retries:
        :type retries: int
        :raises: SocketBlockedException, GroupingValueMissing
        :return:
        """
        try:
            socket_addresses = self.get_grouper(self.send_grouper)(
                self.send_socks.keys()
            ).get_destinations_for(message.grouping_value)
        except GroupingValueMissing:
            raise GroupingValueMissing("Message '%s' provided an invalid grouping_value: '%s'" % (message.content, message.grouping_value))
        for destination in socket_addresses:
            message.send(
                self.send_socks[destination],
                process_id
            )
            if self.controller_sock and self.send_control_messages and control_message:
                for index in xrange(0, retries):
                    try:
                        message.send_control_message(
                            self.controller_sock,
                            time_consumed=time_consumed,
                            process_name=process_id,
                            destination_uuid=self.process_address_to_uuid[destination],
                            destination_endpoint=destination,
                            sender=sender
                        )
                        break
                    except KeyError:  # If destination is not known, lookup fails
                        time.sleep(random.randrange(1, 3))  # random to avoid peak loads when running multiple processes
                    except zmq.Again:  # If connection fails, retry
                        time.sleep(random.randrange(1, 3))
                else:  # if for loop exited cleanly (no break)
                    raise SocketBlockedException("Control Message for process %s could not be sent after %s attempts"
                                                 % (process_id, retries))


class ConnectionMixin(object):
    def set_send_socks(self, connections, output_queue, context):
        for send_conn in connections[output_queue]['streams']:
            if send_conn not in self.send_socks:
                send_sock = context.socket(zmq.PUSH)
                send_sock.connect(send_conn)
                self.send_socks[send_conn] = send_sock
                self.send_grouper = connections[output_queue]['grouping']

        deleted_connections = [connection for connection in self.send_socks.keys() if connection not in connections[output_queue]['streams']]
        for deleted_connection in deleted_connections:
            self.send_socks[deleted_connection].close()
            del self.send_socks[deleted_connection]

    def connection_thread(self, context=None, refresh_connection_stream=None, input_queue=None, output_queue=None,
                          grouper_cls=None, set_controller_sock=True):
        refresh_connection_sock = context.socket(zmq.SUB)
        refresh_connection_sock.connect(refresh_connection_stream)
        refresh_connection_sock.setsockopt(zmq.SUBSCRIBE, '')  # You must subscribe to something, so this means *all*}
        set_timeouts_on_socket(refresh_connection_sock)

        connections = get_connections_block('_update_connections', refresh_connection_sock)

        while not self.receive_port:
            time.sleep(1)

        # Register as consumer of input stream
        update_connection_sock = context.socket(zmq.PUSH)
        update_connection_sock.connect(connections['_update_connections']['streams'][0])
        intersection_connection_info = {
            'streams': {
                input_queue: ['tcp://%s:%s' % (get_ip(), self.receive_port)]
            },
            'meta': {
                'id': self.process_uuid,
                'name': self.process_name,
                'grouping': None if not grouper_cls else grouper_cls.__name__  # Specify how messages sent to this intersection should be grouped
            }
        }
        Message(None, intersection_connection_info).send(update_connection_sock, producer_uuid=self.process_uuid)
        last_send_time = datetime.datetime.now()

        if set_controller_sock:
            connections = get_connections_block('_message_ack', refresh_connection_sock, existing_connections=connections)
            self.controller_sock = context.socket(zmq.PUSH)
            self.controller_sock.connect(connections['_message_ack']['streams'][0])
            set_timeouts_on_socket(self.controller_sock)
        while True:
            try:
                connections = refresh_connection_sock.recv_json()
                if output_queue in connections:
                    self.set_send_socks(connections, output_queue, context)
                self.process_id_to_name = {process['process_id']: process['process_name'] for process in itertools.chain.from_iterable([queue_details['stream_heartbeats'].values() for queue_details in connections.values()])}
                self.process_address_to_uuid = {address: process['process_id'] for address, process in itertools.chain.from_iterable([queue_details['stream_heartbeats'].items() for queue_details in connections.values()])}
            except zmq.Again:
                time.sleep(1)
            if last_send_time + datetime.timedelta(seconds=10) < datetime.datetime.now():
                Message(None, intersection_connection_info).send(update_connection_sock, producer_uuid=self.process_uuid)
                last_send_time = datetime.datetime.now()



from multiprocessing import Process
from setproctitle import setproctitle
import time
import uuid

from motorway.connection import ConnectionIntersection
from motorway.controller import ControllerIntersection
from motorway.grouping import SendToAllGrouper
from motorway.utils import ramp_result_stream_name
import zmq
import logging
from motorway.webserver import WebserverIntersection

logger = logging.getLogger(__name__)


class Pipeline(object):
    def __init__(self, controller_bind_address="0.0.0.0:7007", run_controller=True, run_webserver=True, run_connection_discovery=True):
        self._streams = {}
        self._stream_consumers = {}
        self._processes = []
        self._ramp_result_streams = []
        self.controller_bind_address = "tcp://%s" % controller_bind_address
        self.context = zmq.Context()
        self.run_controller = run_controller
        self.run_webserver = run_webserver
        self.run_connection_discovery = run_connection_discovery

    def definition(self):
        """
        Extend this method in your :class:`motorway.pipeline.Pipeline` subclass, e.g.::

            class WordCountPipeline(Pipeline):
                def definition(self):
                    self.add_ramp(WordRamp, 'sentence')
                    self.add_intersection(SentenceSplitIntersection, 'sentence', 'word', processes=2)
                    self.add_intersection(WordCountIntersection, 'word', 'word_count', grouper_cls=HashRingGrouper, processes=2)
                    self.add_intersection(AggregateIntersection, 'word_count', grouper_cls=HashRingGrouper, processes=1)
        """
        raise NotImplementedError("You must implement a definition() on your pipeline")

    def _add_process(self, cls, process_instances, process_args, input_stream=None, output_stream=None, show_in_ui=True, process_start_number=0):
        for i in range(process_start_number, process_instances + process_start_number):
            process_uuid = uuid.uuid4()
            process_name = "%s-%s" % (cls.__name__, process_uuid.hex)
            kwargs = {
                'process_uuid': process_uuid
            }
            p = Process(
                target=cls.run,
                args=process_args,
                kwargs=kwargs,
                name=process_name
            )
            self._processes.append(p)
            if show_in_ui:
                if output_stream:
                    if output_stream not in self._stream_consumers:
                        self._stream_consumers[output_stream] = {'producers': [], 'consumers': []}
                    self._stream_consumers[output_stream]['producers'].append(process_name)
                if input_stream not in self._stream_consumers:
                    self._stream_consumers[input_stream] = {'producers': [], 'consumers': []}
                self._stream_consumers[input_stream]['consumers'].append(process_name)

    def add_ramp(self, ramp_class, output_stream, processes=1):
        ramp_result_stream = ramp_result_stream_name(ramp_class.__name__)
        self._ramp_result_streams.append((ramp_class.__name__, ramp_result_stream))
        self._add_process(
            ramp_class,
            processes,
            process_args=(
                output_stream,
                self.controller_bind_address,
                self.run_controller
            ),
            output_stream=output_stream,

        )

    def add_intersection(self, intersection_class, input_stream, output_stream=None, processes=1, grouper_cls=None):
        self._add_process(
            intersection_class,
            processes,
            process_args=(
                input_stream,
                output_stream,
                self.controller_bind_address,
                grouper_cls
            ),
            input_stream=input_stream,
            output_stream=output_stream,
        )

    def run(self):
        """
        Execute the entire pipeline in several sub processes.

        """

        logger.info("Starting Pipeline %s!" % self.__class__.__name__)

        setproctitle("data-pipeline: main")

        # User jobs
        self.definition()
        logger.debug("Loaded definition")

        # Controller Transformer
        if self.run_controller:
            self.add_intersection(ControllerIntersection, '_message_ack', '_web_server')

        if self.run_connection_discovery:
            self.add_intersection(ConnectionIntersection, '_update_connections')

        if self.run_webserver:
            self.add_intersection(WebserverIntersection, '_web_server', grouper_cls=SendToAllGrouper)  # all webservers should receive messages

        logger.debug("Running pipeline")
        for process in self._processes:
            process.start()

        try:
            while True:
                for process in self._processes:
                    assert process.is_alive()
                time.sleep(5)
        except Exception:
            raise
        finally:
            self.kill()
        logger.debug("Finished Pipeline!")

    def kill(self):
        for process in self._processes:
            logger.warn("Terminating %s" % process)
            if process.is_alive():
                process.terminate()



from unittest import TestCase
import unittest
from motorway.utils import percentile_from_dict


class UtilsTestCase(TestCase):
    def test_percentile(self):
        test_dict = {0.0: 2, 1.0: 1, 2.0: 1, 3.0: 3, 4.0: 1}
        self.assertEqual(percentile_from_dict(test_dict, 95), 4.0)



import random
from Queue import Queue
import json
from threading import Thread, Lock, Semaphore
import time
import uuid

import datetime
from boto.dynamodb2.exceptions import ItemNotFound, ConditionalCheckFailedException, \
    ProvisionedThroughputExceededException, LimitExceededException
from boto.dynamodb2.fields import HashKey
from boto.dynamodb2.items import Item
from boto.dynamodb2.table import Table
from boto.dynamodb2.types import STRING
from boto.exception import JSONResponseError
from motorway.messages import Message
from motorway.ramp import Ramp
import boto.kinesis
import boto.ec2.cloudwatch
import boto.kinesis.exceptions
import logging


shard_election_logger = logging.getLogger("motorway.contrib.amazon_kinesis.shard_election")


class KinesisRamp(Ramp):
    stream_name = None
    heartbeat_timeout = 30  # Wait 10 seconds for a heartbeat update, or kill it

    def __init__(self, shard_threads_enabled=True, **kwargs):
        super(KinesisRamp, self).__init__(**kwargs)
        self.conn = boto.kinesis.connect_to_region(**self.connection_parameters())
        assert self.stream_name, "Please define attribute stream_name on your KinesisRamp"

        control_table_name = 'pipeline-control-%s' % self.stream_name

        self.worker_id = str(uuid.uuid4())
        self.semaphore = Semaphore()
        self.uncompleted_ids = {}

        if shard_threads_enabled:
            dynamodb_connection = boto.dynamodb2.connect_to_region(**self.connection_parameters())

            table_config = dict(
                connection=dynamodb_connection,
                schema=[
                    HashKey('shard_id', data_type=STRING),
                ],
                throughput={
                    'read': 10,
                    'write': 10,
                }
            )

            try:
                dynamodb_connection.describe_table(control_table_name)
            except JSONResponseError:
                Table.create(
                    control_table_name,
                    **table_config
                )

            self.control_table = Table(
                control_table_name,
                **table_config
            )

            shards = self.conn.describe_stream(self.stream_name)['StreamDescription']['Shards']
            threads = []
            self.insertion_queue = Queue()
            for i, shard in enumerate(shards):
                self.uncompleted_ids[shard['ShardId']] = []
                t = Thread(target=self.process_shard, name="%s-%s" % (self.__class__.__name__, i), args=(shard['ShardId'], ))
                threads.append(t)
                t.start()

    def claim_shard(self, shard_id):
        shard_election_logger.info("Claiming shard %s" % shard_id)
        control_record = self.control_table.get_item(shard_id=shard_id)
        control_record['worker_id'] = self.worker_id
        control_record['heartbeat'] = 0
        try:
            control_record.save()
        except ConditionalCheckFailedException:  # Someone else edited the record
            shard_election_logger.debug("Failed to claim %s to %s" % (shard_id, self.worker_id))
            return False
        return True

    def can_claim_shard(self, shard_id):
        control_record = self.control_table.get_item(shard_id=shard_id)
        heartbeat = control_record['heartbeat']
        time.sleep(self.heartbeat_timeout)
        if heartbeat == self.control_table.get_item(shard_id=shard_id)['heartbeat']:  # TODO: check worker_id as well
            shard_election_logger.debug("Heartbeat remained unchanged for defined time, taking over")
            return True
        else:
            shard_election_logger.debug("Heartbeat changed, continue sleeping")
        # Balance, if possible
        active_workers = {
            self.worker_id: True
        }
        heartbeats = {
            self.worker_id: 0
        }

        shards = list(self.control_table.scan())
        for shard in shards:  # Update active worker cache
            if shard['worker_id'] not in active_workers:
                heartbeats[shard['worker_id']] = shard['heartbeat']
            if heartbeats[shard['worker_id']] == shard['heartbeat']:
                active_workers[shard['worker_id']] = False
            else:
                active_workers[shard['worker_id']] = True
        number_of_active_workers = sum([1 for is_active in active_workers.values() if is_active])
        number_of_shards = len(shards)
        optimal_number_of_shards_per_worker = number_of_shards / number_of_active_workers
        workers = set([shard['worker_id'] for shard in shards])
        shards_per_worker = {worker: sum([1 for shard in shards if shard['worker_id'] == worker]) for worker in workers}
        for shard in shards:
            if shards_per_worker.get(shard['worker_id'], 0) > optimal_number_of_shards_per_worker:
                if shards_per_worker.get(self.worker_id, 0) < optimal_number_of_shards_per_worker:
                    if shard['shard_id'] == shard_id:
                        shard_election_logger.debug("Taking over %s from %s" % (shard_id, shard['worker_id']))
                        return True
        return False

    def process_shard(self, shard_id):
        while True:
            try:
                try:
                    while True:
                        with self.semaphore:
                            if self.can_claim_shard(shard_id):
                                if self.claim_shard(shard_id):
                                    break
                        time.sleep(random.randrange(2, 15))
                except ItemNotFound:
                    control_record = Item(self.control_table, data={
                        'shard_id': shard_id,
                        'checkpoint': 0,
                        'worker_id': self.worker_id,
                        'heartbeat': 0,
                    })
                    control_record.save()


                control_record = self.control_table.get_item(shard_id=shard_id)
                if control_record['checkpoint'] > 0:
                    iterator = self.conn.get_shard_iterator(
                        self.stream_name,
                        shard_id,
                        "AT_SEQUENCE_NUMBER",
                        starting_sequence_number=str(control_record['checkpoint'])
                    )['ShardIterator']
                else:
                    iterator = self.conn.get_shard_iterator(
                        self.stream_name,
                        shard_id,
                        "LATEST",
                    )['ShardIterator']

                cloudwatch = boto.ec2.cloudwatch.connect_to_region(**self.connection_parameters())
                current_minute = lambda: datetime.datetime.now().minute
                minute = None
                while True:
                    control_record = self.control_table.get_item(shard_id=shard_id)
                    if not control_record['worker_id'] == self.worker_id:
                        shard_election_logger.info("Lost shard %s, going back to standby" % shard_id)
                        break
                    control_record['heartbeat'] += 1
                    if len(self.uncompleted_ids[shard_id]):  # Get the "youngest" uncompleted sequence number
                        control_record['checkpoint'] = min(self.uncompleted_ids[shard_id])
                    control_record.save()

                    result = self.conn.get_records(iterator)
                    for record in result['Records']:
                        self.uncompleted_ids[shard_id].append(record['SequenceNumber'])
                        self.insertion_queue.put(record)
                    # Push metrics to CloudWatch
                    iterator = result['NextShardIterator']
                    delay = result['MillisBehindLatest']
                    if minute != current_minute():  # push once per minute to CloudWatch.
                        minute = current_minute()
                        cloudwatch.put_metric_data(
                            'Motorway/Kinesis',
                            'MillisecondsBehind',
                            value=delay,
                            unit='Milliseconds',
                            dimensions={
                                'Stream': self.stream_name,
                                'Shard': shard_id
                            }
                        )
                    time.sleep(1)  # recommended pause between fetches from AWS
            except ConditionalCheckFailedException:
                pass  # we're no longer worker for this shard
            except (ProvisionedThroughputExceededException, LimitExceededException, boto.kinesis.exceptions.ProvisionedThroughputExceededException, boto.kinesis.exceptions.LimitExceededException) as e:
                time.sleep(random.randrange(5, self.heartbeat_timeout))  # back off for a while


    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def next(self):
        msg = self.insertion_queue.get()
        yield Message(msg['SequenceNumber'], json.loads(msg['Data']), grouping_value=msg['PartitionKey'])

    def success(self, _id):
        for shard_id, uncompleted_ids in self.uncompleted_ids.items():
            if _id in uncompleted_ids:
                uncompleted_ids.remove(_id)


import calendar
import logging
from setproctitle import setproctitle
from threading import Thread
import time
import datetime
import uuid
from motorway.decorators import batch_process
from motorway.messages import Message
from motorway.intersection import Intersection
from motorway.utils import percentile_from_dict, set_timeouts_on_socket
from isodate import parse_duration
import zmq


logger = logging.getLogger(__name__)


class ControllerIntersection(Intersection):
    """
    Responsible for keeping track of the state of each message.

    We're using an constant storage algorithm to track indefinite state. The cost of the constant storage is that
    we never know how far it is in the "tree" of intersections it has to pass through. We create a bunch of dictionaries
    to count how many pending messages each intersection has to provide useful stats in the web UI.
    """

    send_control_messages = False

    def __init__(self, stream_consumers=None, **kwargs):
        super(ControllerIntersection, self).__init__(**kwargs)
        self.stream_consumers = stream_consumers or {}
        self.ramp_socks = {}
        self.messages = {}
        self.failed_messages = {}
        self.process_statistics = {}
        self.queue_processes = {}
        self.process_address_to_uuid = {}  # Maps tcp endpoints to human readable names

    def get_default_process_dict(self):
        return {
            'status': 'running',
            'success': 0,
            'failed': 0,
            'processed': 0,
            'waiting': 0,
            'time_taken': datetime.timedelta(seconds=0),
            'avg_time_taken': datetime.timedelta(seconds=0),
            '95_percentile': datetime.timedelta(seconds=0),
            'frequency': {}.copy(),
            'total_frequency': 0,
            'histogram': {
                minute: {
                    'error_count': 0, 'success_count': 0, 'timeout_count': 0, 'processed_count': 0
                }.copy() for minute in range(0, 60)
            }.copy()
        }.copy()


    @batch_process(wait=1, limit=500)
    def process(self, messages):
        for message in messages:
            original_process = message.producer_uuid
            if original_process not in self.process_statistics:
                self.process_statistics[original_process] = self.get_default_process_dict()

            current_process = message.content['process_name']
            if current_process not in self.process_statistics:
                self.process_statistics[current_process] = self.get_default_process_dict()

            destination_process = message.destination_uuid or self.process_address_to_uuid[message.destination_endpoint]  # Usually we should only have the endpoint in the message
            if destination_process not in self.process_statistics:
                self.process_statistics[destination_process] = self.get_default_process_dict()

            # Create message or update with ack value
            if message.ramp_unique_id not in self.messages:  # Message just created
                assert message.producer_uuid, "Producer UUID missing from %s" % message._message()
                self.messages[message.ramp_unique_id] = [
                    message.producer_uuid,
                    message.ack_value,
                    datetime.datetime.now(),
                    destination_process
                ]  # Set the new value to ack value
            elif message.ack_value >= 0:  # Message processed
                self.messages[message.ramp_unique_id][1] ^= message.ack_value  # XOR the existing value
                # Update process information
                if self.messages[message.ramp_unique_id][1] == Message.SUCCESS:
                    original_process = self.messages[message.ramp_unique_id][0]
                    self.process_statistics[original_process]['success'] += 1
                    self.process_statistics[original_process]['histogram'][datetime.datetime.now().minute]['success_count'] += 1
                    self.success(message.ramp_unique_id, original_process)
                    del self.messages[message.ramp_unique_id]
                else:  # Still not finished - update destination!
                    if destination_process != original_process:
                        self.messages[message.ramp_unique_id][3] = destination_process  # Update destination so we can show what is waiting for this process!
            elif message.ack_value == Message.FAIL:
                if message.ramp_unique_id in self.messages:
                    original_process, ack_value, start_time, destination_process = self.messages[message.ramp_unique_id]
                    del self.messages[message.ramp_unique_id]
                self.process_statistics[original_process]['failed'] += 1
                self.process_statistics[original_process]['histogram'][datetime.datetime.now().minute]['error_count'] += 1
                self.fail(message.ramp_unique_id, error_message=message.error_message, process=original_process)
            self.process_statistics[original_process]['processed'] += 1

            # Update statistics
            if 'duration' in message.content:
                time_taken = parse_duration(message.content['duration'])
                rounded_seconds = round(time_taken.total_seconds(), 0)
                self.process_statistics[current_process]['time_taken'] += time_taken
                if ('msg_type' in message.content and message.content['msg_type'] != "new_msg") or ('sender' in message.content and message.content['sender'] == 'ramp'):
                    self.process_statistics[current_process]['histogram'][datetime.datetime.now().minute]['processed_count'] += 1
                    self.process_statistics[current_process]['frequency'][rounded_seconds] = self.process_statistics[current_process]['frequency'].get(rounded_seconds, 0) + 1
                    self.process_statistics[current_process]['total_frequency'] = sum(self.process_statistics[current_process]['frequency'].values())
                    self.process_statistics[current_process]['95_percentile'] = datetime.timedelta(seconds=percentile_from_dict(self.process_statistics[current_process]['frequency'], 95))
                    self.process_statistics[current_process]['avg_time_taken'] = self.process_statistics[current_process]['time_taken'] / sum(self.process_statistics[current_process]['frequency'].values())
        yield  # Hack: This is actually done by self.update() to trigger it even if there are no messages and to reduce messages to 1/s

    def update(self):
        now = datetime.datetime.now()

        # Check message status
        waiting_messages = {}
        for unique_id, lst in self.messages.items():
            original_process, ack_value, start_time, process = lst
            if unique_id in self.failed_messages:
                del self.messages[unique_id]  # This failed somewhere else in the chain and it was notificed already
            elif (now - start_time) > datetime.timedelta(minutes=30):
                del self.messages[unique_id]  # clean up
                self.process_statistics[process]['histogram'][datetime.datetime.now().minute]['timeout_count'] += 1
                self.fail(unique_id, original_process, error_message="Message timed out")
            elif ack_value > 0:
                waiting_messages[process] = waiting_messages.get(process, 0) + 1

        # Update histograms
        for process in self.process_statistics.keys():
            self.process_statistics[process]['histogram'][(now + datetime.timedelta(minutes=1)).minute] = self.get_default_process_dict()['histogram'][0]  # reset next minute
            self.process_statistics[process]['waiting'] = waiting_messages.get(process, 0)

        # Prepare message but copy the dictionaries so we avoid them being changed while we send it over ZMQ
        message = Message("_controller-%s" % uuid.uuid4(), {
            'process_id_to_name': self.process_id_to_name.copy(),
            'process_statistics': self.process_statistics.copy(),
            'stream_consumers': self.stream_consumers.copy(),
            'failed_messages': self.failed_messages.copy(),
        }, grouping_value=str(self.process_uuid))

        if self.send_socks:  # if we have at least one destination, send the updates
            self.send_message(message, self.process_uuid, control_message=False)

    def set_send_socks(self, connections, output_queue, context):
        super(ControllerIntersection, self).set_send_socks(connections, output_queue, context)

        # Override to look for ramps as well, which only is relevant for controller at the moment
        for queue, queue_info in connections.items():
            if '_ramp' in queue:  # Ramp replies
                if queue not in self.ramp_socks or not self.ramp_socks[queue]:
                    self.ramp_socks[queue] = context.socket(zmq.PUSH)
                    self.ramp_socks[queue].connect(queue_info['streams'][0])  # There should always be exactly one stream for a ramp

    def _update_wrapper(self):
        while True:
            self.update()
            time.sleep(1)

    def fail(self, unique_id, process, error_message=""):
        self.failed_messages[unique_id] = (process, error_message)
        if process not in self.ramp_socks and '_ramp' in process:
            logger.warn("%s not in ramp_socks. Had %s" % (process, self.ramp_socks))
        elif process in self.ramp_socks:
            self.ramp_socks[process].send_json({
                'status': 'fail',
                'id': unique_id
            })

    def success(self, unique_id, process):
        if process not in self.ramp_socks and '_ramp' in process:
            logger.warn("%s not in ramp_socks. Had %s" % (process, self.ramp_socks))
        elif process in self.ramp_socks:
            self.ramp_socks[process].send_json({
                'status': 'success',
                'id': unique_id
            })

    def thread_factory(self, *args, **kwargs):
        factories = super(ControllerIntersection, self).thread_factory(*args, **kwargs)
        thread_update_stats_factory = lambda: Thread(target=self._update_wrapper, name="controller-update_stats")

        return factories + [thread_update_stats_factory]


from unittest import TestCase
from boto.dynamodb2.exceptions import ItemNotFound
from motorway.contrib.amazon_kinesis.ramps import KinesisRamp
from mock import patch


class MockDynamoItem(dict):

    def __init__(self, **kwargs):
        super(MockDynamoItem, self).__init__(**kwargs)

    def save(self):
        pass  # This is a dictionary, so it is just updated straight away


class MockControlTable(object):
    def __init__(self, control_table_item_list):
        self.control_table_item_list = control_table_item_list

    def get_item(self, shard_id=None):
        for item in self.control_table_item_list:
            if item['shard_id'] == shard_id:
                return item
        raise ItemNotFound()

    def scan(self):
        for item in self.control_table_item_list:
            yield item

    def _put_item(self, item_data, expects=None):
        pass


class AmazonKinesisTestCase(TestCase):
    def get_kinesis_ramp(self, control_table_item_list=None, shards=None):
        with patch('boto.kinesis.connect_to_region', return_value=None) as mock_method:
            KinesisRamp.stream_name = "_unittest"
            KinesisRamp.heartbeat_timeout = 0  # Let's not care about waiting in this test
            kinesis_ramp = KinesisRamp(shard_threads_enabled=False)
            if not shards:
                shards = []
            if not control_table_item_list:
                control_table_item_list = [MockDynamoItem(
                    shard_id=shard_id,
                    checkpoint=0,
                    worker_id=i,
                    heartbeat=0,
                ) for i, shard_id in enumerate(shards)]
            for shard in shards:
                kinesis_ramp.uncompleted_ids[shard] = []
            kinesis_ramp.control_table = MockControlTable(control_table_item_list)
            return kinesis_ramp

    def test_can_claim_timeout(self):
        kinesis_ramp = self.get_kinesis_ramp(shards=['shard-1', 'shard-2'])
        self.assertTrue(
            kinesis_ramp.can_claim_shard("shard-1")
        )
        self.assertTrue(
            kinesis_ramp.can_claim_shard("shard-2")
        )

    def test_cannot_claim_timeout(self):
        kinesis_ramp = self.get_kinesis_ramp(shards=['shard-1',])
        def change_heartbeat(seconds):
            kinesis_ramp.control_table.get_item(shard_id='shard-1')['heartbeat'] += 1
        with patch('time.sleep', change_heartbeat) as mock_method:
            self.assertFalse(
                kinesis_ramp.can_claim_shard("shard-1")
            )

    def test_rebalance(self):
        control_table_item_list = [
            MockDynamoItem(
                shard_id='shard-1',
                checkpoint=0,
                worker_id=1,
                heartbeat=0,
            ),
            MockDynamoItem(
                shard_id='shard-2',
                checkpoint=0,
                worker_id=1,
                heartbeat=0,
            ),
        ]
        kinesis_ramp = self.get_kinesis_ramp(shards=['shard-1', 'shard-2'], control_table_item_list=control_table_item_list)

        def change_heartbeat(seconds):
            kinesis_ramp.control_table.get_item(shard_id='shard-1')['heartbeat'] += 1
            kinesis_ramp.control_table.get_item(shard_id='shard-1')['checkpoint'] = 1337
        with patch('time.sleep', change_heartbeat) as mock_method:
            self.assertTrue(
                kinesis_ramp.can_claim_shard("shard-1")
            )

    def test_claim_shard(self):
        kinesis_ramp = self.get_kinesis_ramp(shards=['shard-1', 'shard-2'])
        self.assertTrue(
            kinesis_ramp.claim_shard('shard-1')
        )

    def test_checkpoint_transfer(self):
        control_table_item_list = [
            MockDynamoItem(
                shard_id='shard-1',
                checkpoint=0,
                worker_id=1,
                heartbeat=0,
            ),
            MockDynamoItem(
                shard_id='shard-2',
                checkpoint=0,
                worker_id=1,
                heartbeat=0,
            ),
        ]
        kinesis_ramp_a = self.get_kinesis_ramp(shards=['shard-1', 'shard-2'], control_table_item_list=control_table_item_list)
        kinesis_ramp_b = self.get_kinesis_ramp(shards=['shard-1', 'shard-2'], control_table_item_list=control_table_item_list)
        kinesis_ramp_a.claim_shard("shard-1")
        kinesis_ramp_a.claim_shard("shard-2")

        self.assertEqual(kinesis_ramp_b.control_table.get_item('shard-1')['heartbeat'], 0)
        control_table_item_list[0]['checkpoint'] = 1337  # Simulate changing to DynamoDB entry
        kinesis_ramp_b.claim_shard('shard-1')
        self.assertEqual(kinesis_ramp_b.control_table.get_item('shard-1')['worker_id'], kinesis_ramp_b.worker_id)
        self.assertEqual(kinesis_ramp_b.control_table.get_item('shard-1')['checkpoint'], 1337)


import json
import multiprocessing
import traceback
import uuid
import datetime
from isodate import duration_isoformat
from motorway.utils import DateTimeAwareJsonEncoder
import logging


class Message(object):
    """
    :param ramp_unique_id: the unique message ID delivered back upon completion to the ramp
    :param content: any json serializable content
    :param grouping_value: String that can be used for routing messages consistently to the same receiver
    :return:
    """
    FAIL = -1
    SUCCESS = 0

    def __init__(self, ramp_unique_id, content=None, ack_value=None, controller_queue=None, grouping_value=None,
                 error_message=None, process_name=None, producer_uuid=None, destination_endpoint=None,
                 destination_uuid=None):
        self.ramp_unique_id = ramp_unique_id
        self.content = content
        if not ack_value:
            ack_value = uuid.uuid4().int
        self.ack_value = ack_value
        self.controller_queue = controller_queue
        self.grouping_value = grouping_value
        self.error_message = error_message
        self.process_name = process_name
        self.producer_uuid = producer_uuid
        self.destination_endpoint = destination_endpoint
        self.destination_uuid = destination_uuid
        self.init_time = datetime.datetime.now()

    @classmethod
    def new(cls, message, content, grouping_value=None, error_message=None):
        """
        Creates a new message, based on an existing message. This has the consequence that it will be tracked together
            and the tap will not be notified until every message in the chain is properly ack'ed.

        :param message: Message instance, as received by the intersection
        :param content: Any value that can be serialized into json
        :param grouping_value: String that can be used for routing messages consistently to the same receiver
        """
        return cls(ramp_unique_id=message.ramp_unique_id, content=content, grouping_value=grouping_value,
                   error_message=error_message, producer_uuid=message.producer_uuid)

    @classmethod
    def from_message(cls, message, controller_queue, process_name=None):
        """

        :param message: Message dict (converted from JSON)
        :param controller_queue:
        :param process_name: UUID of the process processing this message (as string)
        :return:
        """
        # assert type(message) is dict, "message (%s) should be dict" % message
        if type(message) is not dict:
            logging.error("Expected type dict, got type %s - message: %s", type(message), message)
        message['process_name'] = process_name
        # assert 'producer_uuid' in message, "missing uuid %s" % message
        return cls(controller_queue=controller_queue, **message)

    def _message(self):
        return {
            'content': self.content,
            'ramp_unique_id': self.ramp_unique_id,
            'ack_value': self.ack_value,
            'grouping_value': self.grouping_value,
            'producer_uuid': self.producer_uuid
        }

    def as_json(self):
        return json.dumps(self._message(), cls=DateTimeAwareJsonEncoder)

    def send(self, queue, producer_uuid=None):
        if producer_uuid and not self.producer_uuid:  # Check if provided and we didn't get one already
            self.producer_uuid = producer_uuid
        elif not self.producer_uuid:
            assert self.producer_uuid
        queue.send(
            self.as_json()
        )

    def send_control_message(self, controller_queue, time_consumed=None, process_name=None, destination_endpoint=None,
                             destination_uuid=None, sender=None):
        """
        Control messages are notifications that a new message have been created, so the controller can keep track of
        this particular message and let the ramp know once the entire tree of messages has been completed.

        This is called implicitly on yield Message(_id, 'message')

        :param process_name: UUID of the process processing this message (as string)
        """
        content = {
            'process_name': process_name,
            'msg_type': 'new_msg',
        }
        if not self.producer_uuid:
            raise Exception("Cannot send control message without producer UUID")
        if time_consumed:
            # Ramps provide time consumed, since we don't know the "start time" like in a intersection
            # where it's clear when the message is received and later 'acked' as the last action
            content['duration'] = duration_isoformat(time_consumed)
            content['sender'] = sender
        controller_queue.send_json({
            'ramp_unique_id': self.ramp_unique_id,
            'ack_value': self.ack_value,
            'content': content,
            'producer_uuid': self.producer_uuid,
            'destination_endpoint': destination_endpoint,
            'destination_uuid': destination_uuid
        })

    def ack(self, time_consumed=None):
        """
        Send a message to the controller that this message was properly processed
        """
        self.controller_queue.send_json({
            'ramp_unique_id': self.ramp_unique_id,
            'ack_value': self.ack_value,
            'content': {
                'process_name': self.process_name,
                'msg_type': 'ack',
                'duration': duration_isoformat(time_consumed or (datetime.datetime.now() - self.init_time))
            },
            'producer_uuid': self.producer_uuid,
            'destination_uuid': self.producer_uuid
        })

    def fail(self, error_message="", capture_exception=True):
        """
        Send a message to the controller that this message failed to process
        """
        self.controller_queue.send_json({
            'ramp_unique_id': self.ramp_unique_id,
            'ack_value': -1,
            'content': {
                'process_name': self.process_name,
                'msg_type': 'fail',
                'duration': duration_isoformat(datetime.datetime.now() - self.init_time)
            },
            'producer_uuid': self.producer_uuid,
            'destination_uuid': self.producer_uuid,
            'error_message': error_message if not capture_exception else traceback.format_exc()
        })

    def __repr__(self):
        return "<Message: %s> %s" % (self.ramp_unique_id, self.content)






import json
import boto.kinesis
from motorway.intersection import Intersection


class KinesisInsertIntersection(Intersection):
    stream_name = None

    def __init__(self, **kwargs):
        super(KinesisInsertIntersection, self).__init__(**kwargs)
        self.conn = boto.kinesis.connect_to_region(**self.connection_parameters())
        assert self.stream_name, "Please define attribute stream_name on your KinesisInsertIntersection"

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def process(self, message):
        self.conn.put_record(
            self.stream_name,
            json.dumps(message.content),
            message.grouping_value
        )
        self.ack(message)
        yield



import random
from Queue import Empty
import logging
import multiprocessing
from threading import Thread
import time
from time import time as _time
import uuid
import datetime

from motorway.exceptions import SocketBlockedException
from motorway.messages import Message
from motorway.mixins import GrouperMixin, SendMessageMixin, ConnectionMixin
from motorway.threads import ThreadRunner
from motorway.utils import set_timeouts_on_socket
import zmq

logger = logging.getLogger(__name__)


class Intersection(GrouperMixin, SendMessageMixin, ConnectionMixin, ThreadRunner):
    """
    Intersections receive messages and generate either:

    - A spin-off message

    Spin-off messages will keep track of the state of the entire message tree and re-run it if failed. This means that
    if you want to re-run the message all the way from the ramp in case of an error, you should make a spin-off message.

    Message.new(message, {
        {
            'word': 'hello',
            'count': 1
        },
        grouping_value='hello'
    })

    - A brand new message

    The message will be created with the intersection as producer. The intersection will not receive feedback if it
    is successful or not and hence will not be re-tried in the case of an error.

    Message(uuid.uuid4()
    """

    send_control_messages = True

    def __init__(self, process_uuid=None):
        super(Intersection, self).__init__()
        self.messages_processed = 0
        self.process_uuid = process_uuid.hex if process_uuid else uuid.uuid4().hex
        self.process_name = multiprocessing.current_process().name
        self.receive_port = None
        self.send_socks = {}
        self.send_grouper = None
        self.controller_sock = None
        self.message_batch_start = datetime.datetime.now()  # This is used to time how much time messages take
        self.process_id_to_name = {}  # Maps UUIDs to human readable names
        self.process_address_to_uuid = {}

    def thread_factory(self, input_stream, output_stream=None, refresh_connection_stream=None, grouper_cls=None, process_uuid=None):
        context = zmq.Context()

        # Create Thread Factories :-)

        thread_update_connections_factory = lambda: Thread(target=self.connection_thread, name="connection_thread", kwargs={
            'refresh_connection_stream': refresh_connection_stream,
            'context': context,
            'input_queue': input_stream,
            'output_queue': output_stream,
            'grouper_cls': grouper_cls
        })

        thread_main_factory = lambda: Thread(target=self.receive_messages, name="message_producer", kwargs={
            'context': context,
            'output_stream': output_stream,
            'grouper_cls': grouper_cls,
        })

        return [thread_update_connections_factory, thread_main_factory]

    def _process(self, receive_sock, controller_sock=None):
        try:
            if getattr(self.process, 'batch_process', None):
                poller = zmq.Poller()
                poller.register(receive_sock, zmq.POLLIN)
                value = []
                end_time = _time() + self.process.wait
                while end_time > _time() and len(value) < self.process.limit:
                    socks = dict(poller.poll(timeout=1000))
                    if socks.get(receive_sock) == zmq.POLLIN:
                        value.append(receive_sock.recv_json())
                message_count = len(value)
            else:
                poller = zmq.Poller()
                poller.register(receive_sock, zmq.POLLIN)
                socks = dict(poller.poll(timeout=1000))
                if socks.get(receive_sock) == zmq.POLLIN:
                    value = receive_sock.recv_json()
                else:
                    return
                message_count = 1
            if value:
                self.messages_processed += message_count
                if isinstance(value, list):
                    message = [Message.from_message(m, controller_sock, process_name=self.process_uuid) for m in value]
                else:
                    message = Message.from_message(value, controller_sock, process_name=self.process_uuid)
                try:
                    self.message_batch_start = datetime.datetime.now()
                    for generated_message in self.process(message):
                        if generated_message is not None and self.send_socks:
                            self.send_message(generated_message, self.process_uuid, time_consumed=(datetime.datetime.now() - self.message_batch_start), control_message=self.send_control_messages)
                            self.message_batch_start = datetime.datetime.now()
                except Exception as e:
                    logger.error(str(e), exc_info=True)

                    # Don't send control messages if e.g. web server or other system process
                    if self.send_control_messages:
                        if isinstance(message, list):
                            [m.fail() for m in message]
                        else:
                            message.fail()

        except Empty:  # Didn't receive anything from ZMQ
            pass

    def ack(self, message, retries=100):
        for index in xrange(0, retries):
            try:
                message.ack(time_consumed=(datetime.datetime.now() - self.message_batch_start))
                break
            except zmq.Again:
                time.sleep(random.randrange(1, 3))  # to avoid peak loads when running multiple processes
        else:  # if for loop exited cleanly (no break)
            raise SocketBlockedException("Acknowledge for process %s could not be sent after %s attempts" % (self.process_name, retries))

        self.message_batch_start = datetime.datetime.now()

    def fail(self, message, **kwargs):
        message.fail(**kwargs)

    def process(self, message):
        """
        This function is called continuously by the intersection.

        :yield: :class:`motorway.messages.Message` instance
        :param message: :class:`motorway.messages.Message` instance or :func:`list` if using
            :func:`motorway.decorators.batch_process`

        """
        raise NotImplementedError()

    def receive_messages(self, context=None, output_stream=None, grouper_cls=None):
        """
        Continously read and process using _process function

        """
        receive_sock = context.socket(zmq.PULL)
        self.receive_port = receive_sock.bind_to_random_port("tcp://*")
        set_timeouts_on_socket(receive_sock)

        if self.send_control_messages:
            while not self.controller_sock:
                time.sleep(1)

        while True:
            self._process(receive_sock, controller_sock=self.controller_sock)



import random
from hash_ring import HashRing
import zmq
import logging

logger = logging.getLogger(__name__)


class GroupingValueMissing(Exception):
    pass


class BaseGrouper(object):
    def __init__(self, destinations):
        self.destinations = destinations

    def get_destinations_for(self, value):
        raise NotImplementedError("Please implement get_destinations_for")


class HashRingGrouper(BaseGrouper):
    """
    Group messages based on a consistent hashing algorithm. e.g. a message with the same grouping key will always
    go to the same intersection
    """
    def __init__(self, *args, **kwargs):
        super(HashRingGrouper, self).__init__(*args, **kwargs)
        self.hash_ring = HashRing(self.destinations)

    def get_destinations_for(self, value):
        if type(value) is int:
            value = str(value)
        try:
            return [self.hash_ring.get_node(value)]
        except TypeError:
            raise GroupingValueMissing("'%s' is an invalid grouping value for HashRingGrouper" % (value, ))


class RandomGrouper(BaseGrouper):
    """
    Select a random intersection to receive the message
    """
    def get_destinations_for(self, value):
        return [random.choice(self.destinations)]


class SendToAllGrouper(BaseGrouper):
    """
    Send messages to all intersections
    """
    def get_destinations_for(self, value):
        return self.destinations



import logging.config
from motorway.pipeline import Pipeline

logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': "%(log_color)s%(levelname)-8s%(reset)s %(name)-32s %(processName)-32s %(blue)s%(message)s"
        }
    },
    'loggers': {
        'motorway': {
            'level': 'WARN',
            'handlers': ['console'],
            'propagate': False
        },
        'werkzeug': {
            'level': 'WARN',
            'handlers': ['console'],
            'propagate': False,
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'colored'
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }

})


class WebPipeline(Pipeline):
    def definition(self):
        pass


WebPipeline(run_controller=False, run_webserver=True, run_connection_discovery=False, controller_bind_address="connections:7007").run()


from collections import defaultdict
from motorway.messages import Message
from motorway.pipeline import Pipeline
from motorway.ramp import Ramp
from motorway.intersection import Intersection


class WordRamp(Ramp):
    sentences = [
        "Oak is strong and also gives shade.",
        "Cats and dogs each hate the other.",
        "The pipe began to rust while new.",
        "Open the crate but don't break the glass.",
        "Add the sum to the product of these three.",
        "Thieves who rob friends deserve jail.",
        "The ripe taste of cheese improves with age.",
        "Act on these orders with great speed.",
        "The hog crawled under the high fence.",
        "Move the vat over the hot fire.",
    ]

    def __init__(self):
        super(WordRamp, self).__init__()
        self.completed = False
        self.successes = []
        self.errors = []

    def next(self):
        if not self.completed:
            for _id, sentence in enumerate(self.sentences):
                yield Message(_id, sentence)
            self.completed = True

    def success(self, _id):
        self.successes.append(_id)

    def failed(self, _id):
        self.errors.append(_id)


class SentenceSplitIntersection(Intersection):
    def process(self, message):
        for word in message.content.split(" "):
            yield Message.new(message, word, grouping_value=word)
        self.ack(message)


class WordCountIntersection(Intersection):
    def __init__(self):
        super(WordCountIntersection, self).__init__()
        self._count = defaultdict(int)

    def process(self, messages):
        for message in messages:
            self._count[message.content] += 1
            self.ack(message)


class SampleWordCountPipeline(Pipeline):
    def definition(self):
        self.add_ramp(WordRamp, 'sentence')
        self.add_intersection(SentenceSplitIntersection, 'sentence', 'word', processes=1)
        self.add_intersection(WordCountIntersection, 'word',)


import json
from boto.sqs.message import Message


class SQSJSONMessage(Message):
    def encode(self, value):
        return json.dumps(value)

    def decode(self, value):
        return json.loads(value)


import mock


class ZMQSockMock(object):
    def __init__(self, control_messages):
        self.control_messages = control_messages

    def send_json(self, message):
        self.control_messages.append(message)


def override_process_name(new_name):
    return mock.patch("multiprocessing.current_process", new=type("mock", (object, ), {'name': new_name}))


from motorway.contrib.amazon_sqs.utils import SQSJSONMessage
from motorway.messages import Message
from motorway.ramp import Ramp
import boto.sqs


class SQSRamp(Ramp):
    queue_name = None
    sqs_message_class = SQSJSONMessage
    json_group_key = None

    def __init__(self, *args, **kwargs):
        super(SQSRamp, self).__init__(*args, **kwargs)
        conn = boto.sqs.connect_to_region(**self.connection_parameters())
        assert self.queue_name, "Please define attribute queue_name on your SQSRamp"
        self.queue = conn.create_queue(self.queue_name)
        self.queue.set_message_class(self.sqs_message_class)
        self.messages = {}

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def next(self):
        for msg in self.queue.get_messages(num_messages=10, wait_time_seconds=5, visibility_timeout=10*60):
            # Gets max 10 messages, waiting for max 5 seconds to receive them and blocks other from receiving it for 10m
            self.messages[msg.id] = msg
            if self.json_group_key:
                body = msg.get_body()
                yield Message(msg.id, body, grouping_value=body[self.json_group_key])
            else:
                yield Message(msg.id, msg.get_body())

    def success(self, _id):
        self.queue.delete_message(self.messages[_id])  # TODO: Do this on ack


import logging
import multiprocessing
from setproctitle import setproctitle
import datetime
from threading import Thread
import uuid
from motorway.grouping import GroupingValueMissing
from motorway.messages import Message
from motorway.mixins import GrouperMixin, SendMessageMixin, ConnectionMixin
from motorway.utils import set_timeouts_on_socket, get_connections_block
import zmq
import time
import random


logger = logging.getLogger(__name__)


class Ramp(GrouperMixin, SendMessageMixin, ConnectionMixin, object):
    """
    All messages must at some point start at a ramp, which ingests data into the pipeline from an external system or
    generates data it self (such as random words in the tutorial)
    """

    send_control_messages = True

    def __init__(self, runs_on_controller=False, process_uuid=None):
        super(Ramp, self).__init__()
        self.runs_on_controller = runs_on_controller
        self.send_socks = {}
        self.send_grouper = None
        self.receive_port = None
        self.process_name = multiprocessing.current_process().name
        self.process_uuid = "_ramp-%s" % process_uuid

    def next(self):
        """
        This function is called continuously by the ramp.

        .. WARNING::
            Do not block this for a long time or make a while True loop inside it. Betwen every :func:`motorway.ramp.Ramp.next` run,
            some essential operations are run, including receiving acks from the :class:`motorway.controller.Controller`

        :yield: :class:`motorway.messages.Message` instance
        """
        raise NotImplementedError()

    def __iter__(self):
        yield self.next()

    def success(self, _id):
        """
        Called when a message was successfully ack'ed through the entire pipeline.

        :param _id: The id of the message that was successful
        """
        pass

    def failed(self, _id):
        """
        Called when a message failed somewhere in the pipeline. The message might not be entirely finished processing
        at this point and this function might be called multiple times.

        :param _id: The id of the message that failed
        """
        pass

    @classmethod
    def run(cls, queue, refresh_connection_stream=None, runs_on_controller=False, process_uuid=None):

        self = cls(runs_on_controller=runs_on_controller, process_uuid=process_uuid)

        logger.info("Running %s" % self.process_name)
        logger.debug("%s pushing to %s" % (self.process_name, queue))
        setproctitle("data-pipeline: %s" % self.process_name, )

        context = zmq.Context()

        # Create Thread Factories :-)

        thread_update_connections_factory = lambda: Thread(target=self.connection_thread, name="connection_thread", kwargs={
            'refresh_connection_stream': refresh_connection_stream,
            'context': context,
            'input_queue': self.process_uuid,
            'output_queue': queue,
        })
        #self, context=None, refresh_connection_stream=None, input_queue=None, output_queue=None, process_id=None,
                          #process_name=None, grouper_cls=None, set_controller_sock=True

        thread_main_factory = lambda: Thread(target=self.message_producer, name="message_producer", kwargs={
            'context': context,
            'process_id': self.process_uuid,
        })

        thread_results_factory = lambda: Thread(target=self.receive_replies, name="results", kwargs={
            'context': context,
        })

        # Run threads

        thread_update_connections = thread_update_connections_factory()
        thread_update_connections.start()

        thread_main = thread_main_factory()
        thread_main.start()

        thread_results = thread_results_factory()
        thread_results.start()

        while True:
            if not thread_update_connections.isAlive():
                logger.error("Thread thread_update_connections crashed in %s" % self.__class__.__name__)
                thread_update_connections = thread_update_connections_factory()
                thread_update_connections.start()
            if not thread_main.isAlive():
                logger.error("Thread thread_main crashed in %s" % self.__class__.__name__)
                thread_main = thread_main_factory()
                thread_main.start()
            if not thread_results.isAlive():
                logger.error("Thread thread_results crashed in %s" % self.__class__.__name__)
                thread_results = thread_results_factory()
                thread_results.start()
            time.sleep(10)

    def receive_replies(self, context=None):
        result_sock = context.socket(zmq.PULL)
        self.receive_port = result_sock.bind_to_random_port("tcp://*")
        logger.debug("Result port is %s" % self.receive_port)
        set_timeouts_on_socket(result_sock)

        while True:
            try:
                message_reply = result_sock.recv_json()
                if message_reply['status'] == 'success':
                    self.success(message_reply['id'])
                elif message_reply['status'] == 'fail':
                    self.failed(message_reply['id'])
                else:
                    logger.warn("Received unknown status feedback %s" % message_reply['status'])
            except zmq.Again:
                time.sleep(1)

    def should_run(self):
        """
        Subclass to define rules whether this tap should run or not. Mainly used for ensuring a tap only runs once
         across the network

        :return: bool
        """
        return self.runs_on_controller

    def message_producer(self, context=None, process_id=None):

        while True:
            if not self.send_socks:
                logger.debug("Waiting for send_socks")
                time.sleep(1)
            elif self.should_run():
                start_time = datetime.datetime.now()
                try:
                    for received_message_result in self:
                        for generated_message in received_message_result:
                            if generated_message is not None:
                                self.send_message(generated_message, process_id, time_consumed=datetime.datetime.now() - start_time, sender='ramp')
                            start_time = datetime.datetime.now()
                except Exception as e:  # Never let user code crash the whole thing!!
                    logger.exception(e)
